##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import z3c.pagelet.zcml
import zope.component
import zope.configuration.fields
import zope.interface
import zope.schema
import zope.security.checker
import zope.security.zcml
from zope.browserpage.metaconfigure import _handle_allowed_attributes
from zope.browserpage.metaconfigure import _handle_allowed_interface
from zope.browserpage.metaconfigure import _handle_for
from zope.browserpage.metaconfigure import _handle_permission
from zope.configuration.exceptions import ConfigurationError
from zope.publisher.interfaces.browser import IDefaultBrowserLayer

from z3c.wizard import interfaces
from z3c.wizard import step
from z3c.wizard import wizard


class IWizardDirective(z3c.pagelet.zcml.IPageletDirective):
    """A directive to register a new wizard.

    The wizard directive also supports an undefined set of keyword arguments
    that are set as attributes on the wizard after creation.
    """


class IWizardStepDirective(zope.interface.Interface):
    """A directive to register a new wizard step.

    The wizard step directive also supports an undefined set of keyword
    arguments that are set as attributes on the wizard step after creation.
    """

    name = zope.schema.TextLine(
        title="The name of the pagelet.",
        description="The name shows up in URLs/paths. For example 'foo'.",
        required=True)

    class_ = zope.configuration.fields.GlobalObject(
        title="Class",
        description="A class that provides attributes used by the pagelet.",
        required=True)

    permission = zope.security.zcml.Permission(
        title="Permission",
        description="The permission needed to use the pagelet.",
        required=True)

    for_ = zope.configuration.fields.GlobalObject(
        title="Context",
        description="The content interface or class this pagelet is for.",
        required=False)

    layer = zope.configuration.fields.GlobalInterface(
        title="The layer the view is in.",
        description="""
        A skin is composed of layers. It is common to put skin
        specific views in a layer named after the skin. If the 'layer'
        attribute is not supplied, it defaults to 'default'.""",
        required=False)

    wizard = zope.configuration.fields.GlobalObject(
        title="Wizard",
        description="The wizard interface or class this step is for.",
        required=False)

    provides = zope.configuration.fields.GlobalInterface(
        title="The interface this pagelets provides.",
        description="""
        A pagelet can provide an interface.  This would be used for
        views that support other views.""",
        required=False,
        default=interfaces.IPagelet)

    allowed_interface = zope.configuration.fields.Tokens(
        title="Interface that is also allowed if user has permission.",
        description="""
        By default, 'permission' only applies to viewing the view and
        any possible sub views. By specifying this attribute, you can
        make the permission also apply to everything described in the
        supplied interface.

        Multiple interfaces can be provided, separated by
        whitespace.""",
        required=False,
        value_type=zope.configuration.fields.GlobalInterface())

    allowed_attributes = zope.configuration.fields.Tokens(
        title="View attributes that are also allowed if the user"
              " has permission.",
        description="""
        By default, 'permission' only applies to viewing the view and
        any possible sub views. By specifying 'allowed_attributes',
        you can make the permission also apply to the extra attributes
        on the view object.""",
        required=False,
        value_type=zope.configuration.fields.PythonIdentifier())


# Arbitrary keys and values are allowed to be passed to the wizard.
IWizardDirective.setTaggedValue('keyword_arguments', True)


# Arbitrary keys and values are allowed to be passed to the step.
IWizardStepDirective.setTaggedValue('keyword_arguments', True)


# wizard directive
def wizardDirective(
        _context, class_, name, permission, for_=zope.interface.Interface,
        layer=IDefaultBrowserLayer, provides=interfaces.IWizard,
        allowed_interface=None, allowed_attributes=None, **kwargs):

    # Security map dictionary
    required = {}

    # Get the permission; mainly to correctly handle CheckerPublic.
    permission = _handle_permission(_context, permission)

    # The class must be specified.
    if not class_:
        raise ConfigurationError("Must specify a class.")

    if not zope.interface.interfaces.IInterface.providedBy(provides):
        raise ConfigurationError("Provides interface provide IInterface.")

    ifaces = list(zope.interface.Declaration(provides).flattened())
    if interfaces.IWizard not in ifaces:
        raise ConfigurationError("Provides interface must inherit IWizard.")

    # Build a new class that we can use different permission settings if we
    # use the class more then once.
    cdict = {}
    cdict['__name__'] = name
    cdict.update(kwargs)
    new_class = type(class_.__name__, (class_, wizard.Wizard), cdict)

    # Set up permission mapping for various accessible attributes
    _handle_allowed_interface(
        _context, allowed_interface, permission, required)
    _handle_allowed_attributes(
        _context, allowed_attributes, permission, required)
    _handle_allowed_attributes(
        _context, kwargs.keys(), permission, required)
    _handle_allowed_attributes(
        _context, ('__call__', 'browserDefault', 'update', 'render',
                   'publishTraverse'), permission, required)

    # Register the interfaces.
    _handle_for(_context, for_)

    # provide the custom provides interface if not allready provided
    if not provides.implementedBy(new_class):
        zope.interface.classImplements(new_class, provides)

    # Create the security checker for the new class
    zope.security.checker.defineChecker(
        new_class, zope.security.checker.Checker(required))

    # register pagelet
    _context.action(
        discriminator=('pagelet', for_, layer, name),
        callable=zope.component.zcml.handler,
        args=('registerAdapter',
              new_class, (for_, layer), provides, name, _context.info),)


def wizardStepDirective(
        _context, class_, name, permission, for_=zope.interface.Interface,
        layer=IDefaultBrowserLayer, wizard=interfaces.IWizard,
        provides=interfaces.IStep, allowed_interface=None,
        allowed_attributes=None, **kwargs):

    # Security map dictionary
    required = {}

    # Get the permission; mainly to correctly handle CheckerPublic.
    permission = _handle_permission(_context, permission)

    # The class must be specified.
    if not class_:
        raise ConfigurationError("Must specify a class.")

    if not zope.interface.interfaces.IInterface.providedBy(provides):
        raise ConfigurationError("Provides interface provide IInterface.")

    ifaces = list(zope.interface.Declaration(provides).flattened())
    if interfaces.IPagelet not in ifaces:
        raise ConfigurationError("Provides interface must inherit IPagelet.")

    if not interfaces.IWizard.implementedBy(wizard):
        raise ConfigurationError("Provides interface must inherit IWizard.")

    # Build a new class that we can use different permission settings if we
    # use the class more then once.
    cdict = {}
    cdict['__name__'] = name
    cdict.update(kwargs)
    new_class = type(class_.__name__, (class_, step.Step), cdict)

    # Set up permission mapping for various accessible attributes
    _handle_allowed_interface(
        _context, allowed_interface, permission, required)
    _handle_allowed_attributes(
        _context, allowed_attributes, permission, required)
    _handle_allowed_attributes(
        _context, kwargs.keys(), permission, required)
    _handle_allowed_attributes(
        _context, ('__call__', 'browserDefault', 'update', 'render',
                   'publishTraverse'), permission, required)

    # Register the interfaces.
    _handle_for(_context, for_)

    # provide the custom provides interface if not allready provided
    if not provides.implementedBy(new_class):
        zope.interface.classImplements(new_class, provides)

    # Create the security checker for the new class
    zope.security.checker.defineChecker(
        new_class, zope.security.checker.Checker(required))

    # register pagelet
    _context.action(
        discriminator=('pagelet', for_, layer, name),
        callable=zope.component.zcml.handler,
        args=('registerAdapter',
              new_class, (for_, layer, wizard), provides, name, _context.info))



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Wizard button actions implementation."""

import doctest
import unittest

import zope.interface
from zope.interface.verify import verifyClass
from zope.interface.verify import verifyObject
from zope.publisher.browser import TestRequest

from z3c.wizard import interfaces
from z3c.wizard import step
from z3c.wizard import testing
from z3c.wizard import wizard


class IContentStub(zope.interface.Interface):
    """Content stub marker."""


@zope.interface.implementer(IContentStub)
class ContentStub:
    """Content stub."""

    def values(self):
        pass


class StepTestClass(step.Step):
    """Simple test step."""


class WizardTestClass(wizard.Wizard):
    """Wizard Test class providing a step."""

    baseURL = '#'

    def __init__(self, context, request):
        super().__init__(context, request)
        self.step = StepTestClass(context, request, self)
        self.step.__name__ = 'first'
        self.step.__parent__ = self


def setStubs():
    zope.component.provideAdapter(
        StepTestClass,
        (IContentStub, None, None), provides=interfaces.IStep, name='first')
    zope.component.provideAdapter(
        StepTestClass,
        (IContentStub, None, None), provides=interfaces.IStep, name='last')


class TestStep(unittest.TestCase):

    def setUp(self):
        setStubs()

    def test_verifyClass(self):
        self.assertTrue(verifyClass(interfaces.IStep, step.Step))

    def test_verifyObject(self):
        content = ContentStub()
        request = TestRequest()
        wiz = wizard.Wizard(content, request)
        step_inst = step.Step(content, request, wiz)
        self.assertTrue(verifyObject(interfaces.IStep, step_inst))


class TestWizard(unittest.TestCase):

    def setUp(self):
        setStubs()

    def test_verifyClass(self):
        self.assertTrue(verifyClass(interfaces.IWizard, wizard.Wizard))

    def test_verifyObject(self):
        wiz = WizardTestClass(ContentStub(), TestRequest())
        self.assertTrue(verifyObject(interfaces.IWizard, wiz))


def test_suite():
    return unittest.TestSuite((
        doctest.DocFileSuite(
            'README.rst',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
        ),
        doctest.DocFileSuite(
            'zcml.rst',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,),
        unittest.defaultTestLoader.loadTestsFromTestCase(TestStep),
        unittest.defaultTestLoader.loadTestsFromTestCase(TestWizard),
    ))



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import zope.i18nmessageid
import zope.interface
import zope.location.interfaces
import zope.schema
from z3c.form import button
from z3c.form import interfaces
from z3c.pagelet.interfaces import IPagelet


_ = zope.i18nmessageid.MessageFactory('z3c')


class IBackButton(interfaces.IButton):
    """A button that redirects to the previous step."""


class INextButton(interfaces.IButton):
    """A button that redirects to the next step."""


class IWizardButtons(zope.interface.Interface):
    """Wizard button interfaces."""

    back = button.Button(
        title=_('Back'),
        condition=lambda form: form.showBackButton)
    zope.interface.alsoProvides(back, (IBackButton,))

    next = button.Button(
        title=_('Next'),
        condition=lambda form: form.showNextButton)
    zope.interface.alsoProvides(next, (INextButton,))

    complete = button.Button(
        title=_('Complete'),
        condition=lambda form: form.showCompleteButton)
    zope.interface.alsoProvides(complete, (INextButton,))


class IStep(interfaces.IForm, IPagelet):
    """An interface marking a step sub-form."""

    available = zope.schema.Bool(
        title='Available',
        description='Marker for available step',
        default=True,
        required=False)

    visible = zope.schema.Bool(
        title='Show step in wizard step menu',
        description='Show step in wizard step menu',
        default=True,
        required=False)

    showRequired = zope.schema.Bool(
        title='Show required label',
        description='Show required label',
        default=True,
        required=False)

    weight = zope.schema.Int(
        title='Step weight in wizard',
        description='Step weight in wizard',
        default=0,
        required=False)

    completed = zope.schema.Bool(
        title='Completed',
        description='Marker for completed step',
        default=False,
        required=False)

    handleApplyOnBack = zope.schema.Bool(
        title='Handle apply changes on back',
        description='Handle apply changes on back will force validation',
        default=False,
        required=False)

    handleApplyOnNext = zope.schema.Bool(
        title='Handle apply changes on next',
        description='Handle apply changes on next will force validation',
        default=True,
        required=False)

    handleApplyOnComplete = zope.schema.Bool(
        title='Handle apply changes on complete',
        description='Handle apply changes on complete will force validation',
        default=True,
        required=False)

    showSaveButton = zope.schema.Bool(
        title='Show save button',
        description='Show save button',
        default=True,
        required=False)

    showBackButton = zope.schema.Bool(
        title='Show back button',
        description='Back button condition',
        default=True,
        required=False)

    showNextButton = zope.schema.Bool(
        title='Show next button',
        description='Next button condition',
        default=True,
        required=False)

    showCompleteButton = zope.schema.Bool(
        title='Show complete button',
        description='Complete button condition',
        default=True,
        required=False)

    def goToStep(stepName):
        """Redirect to step by name."""

    def goToNext():
        """Redirect to next step."""

    def goToBack():
        """Redirect to back step."""

    def applyChanges(data):
        """Generic form save method taken from z3c.form.form.EditForm."""

    def doHandleApply(action):
        """Extract data and calls applyChanges."""

    def doBack(action):
        """Process back action and return True on sucess."""

    def doNext(action):
        """Process next action and return True on sucess."""

    def doComplete(action):
        """Process complete action and return True on sucess."""

    def update():
        """Update the step."""

    def render():
        """Render the step content w/o wrapped layout."""

    def __call__():
        """Compute a response body including the layout"""


class IWizard(zope.location.interfaces.ILocation):
    """An interface marking the controlling wizard form."""

    firstStepAsDefault = zope.schema.Bool(
        title='Show first step as default',
        description='Show first step or first not completed step as default',
        default=True,
        required=True)

    adjustStep = zope.schema.Bool(
        title='Adjust step',
        description='Force fallback (redirect) to last incomplete step',
        default=True,
        required=False)

    confirmationPageName = zope.schema.ASCIILine(
        title='Confirmation page name',
        description='The confirmation page name shown after completed',
        default=None,
        required=False)

    cssActive = zope.schema.ASCIILine(
        title='Active step menu CSS class',
        description='The active step menu CSS class',
        default='selected',
        required=False)

    cssInActive = zope.schema.ASCIILine(
        title='In-Active step menu item CSS class',
        description='The in-active step menu item CSS class',
        default=None,
        required=False)

    stepInterface = zope.interface.Attribute('Step lookup interface.')

    steps = zope.interface.Attribute(
        """List of one or more IStep (can be lazy).""")

    stepMenu = zope.interface.Attribute("""Step menu info.""")

    step = zope.schema.Object(
        title='Current step',
        description='Current step',
        schema=IStep)

    completed = zope.schema.Bool(
        title='Completed',
        description='Marker for completed step',
        default=False,
        required=False)

    isFirstStep = zope.schema.Bool(
        title='Is first step',
        description='Is first step',
        default=False,
        required=False)

    isLastStep = zope.schema.Bool(
        title='Is last step',
        description='Is last step',
        default=False,
        required=False)

    previousStepName = zope.schema.TextLine(
        title='Previous step name',
        description='Previous step name',
        default=None,
        required=False)

    nextStepName = zope.schema.TextLine(
        title='Next step name',
        description='Next step name',
        default=None,
        required=False)

    def doAdjustStep():
        """Ensure traversal is only possible up to first not completed step.

        Trying to traverse to another step redirects to the first incomplete
        one.
        """

    def getDefaultStep():
        """Can return the first or first not completed step as default."""

    def updateActions():
        """Update wizard actions."""

    def publishTraverse(request, name):
        """Traverse to step by it's name."""

    def browserDefault(request):
        """The default step is our browserDefault traversal setp."""

    def goToStep(stepName):
        """Redirect to the step by name."""

    def goToBack():
        """Redirect to next step if previous get sucessfuly processed."""

    def goToNext():
        """Redirect to next step if previous get sucessfuly processed."""

    def doBack(action):
        """Process something if back action get exceuted."""

    def doNext(action):
        """Process something if next action get exceuted."""

    def doComplete(action):
        """Process something if complete action get exceuted."""

    def doFinish():
        """Process something on complete wizard."""

    def update():
        """Adjust step and update actions."""



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import zope.component
import zope.event
import zope.interface
import zope.lifecycleevent
from z3c.form import button
from z3c.form.interfaces import IDataManager
from z3c.formui import form

from z3c.wizard import interfaces
from z3c.wizard.interfaces import _


def addStep(self, name, label=None, weight=None, available=None, **kws):
    step = zope.component.getMultiAdapter((self.context, self.request, self),
                                          interfaces.IStep, name=name)
    step.__name__ = name
    if label is not None:
        step.label = label
    if weight is not None:
        step.weight = weight
    if available is not None:
        step.available = available
    for name, value in kws.items():
        setattr(step, name, value)
    return step


@zope.interface.implementer(interfaces.IStep)
class Step(form.Form):
    """Wizard base step implementation.

    The step offers hooks for action handlers for all wizard actions. The step
    can also provide own actions and handlers. This actions get rendered as
    step actions. Between the back an next wizard actions.

    A step can access the context or any object which you return by the
    getContent method. See z3c.form for more info about that. If you need a
    complexer wizard setup, you probably have to use a session and store
    temporary collected values in the session and store it if the wizard will
    call doComplete on the last step or in the wizard itself. Such a session
    is not a part of this implementation. This wizard implementation works
    on any context like other z3c.form IForm implementations. For more infos
    see z3c.form which this wizard is based on.
    """

    label = None
    available = True
    visible = True
    weight = 0
    showRequired = True

    handleApplyOnBack = False
    handleApplyOnNext = True
    handleApplyOnComplete = True

    # button condition
    showSaveButton = True

    formErrorsMessage = _('There were some errors.')
    successMessage = _('Data successfully updated.')
    noChangesMessage = _('No changes were applied.')

    def __init__(self, context, request, wizard):
        self.context = context
        self.request = request
        self.wizard = self.__parent__ = wizard

    @property
    def showBackButton(self):
        """Back button condition."""
        return not self.wizard.isFirstStep

    @property
    def showNextButton(self):
        """Next button condition."""
        return not self.wizard.isLastStep

    @property
    def showCompleteButton(self):
        """Complete button condition."""
        return self.wizard.isLastStep and self.wizard.completed

    @property
    def nextURL(self):
        """Next step URL known by wizard."""
        return self.wizard.nextURL

    @property
    def completed(self):
        """Simple default check for find out if a step is complete.

        This method will ensure that we store at least all required form values
        and that this values are valid. You can implement any other or
        additional condition in your custom step implementation.
        """
        content = self.getContent()
        for field in self.fields.values():
            if not field.field.required:
                continue
            dm = zope.component.getMultiAdapter(
                (content, field.field), IDataManager)
            if dm.query(
                    field.field.missing_value) is field.field.missing_value:
                return False
        return True

    def goToStep(self, stepName):
        self.wizard.goToStep(stepName)

    def goToNext(self):
        self.wizard.goToNext()

    def goToBack(self):
        self.wizard.goToBack()

    def applyChanges(self, data):
        """Generic form save method taken from z3c.form.form.EditForm."""
        content = self.getContent()
        changes = form.applyChanges(self, content, data)
        # ``changes`` is a dictionary; if empty, there were no changes
        if changes:
            # Construct change-descriptions for the object-modified event
            descriptions = []
            for interface, names in changes.items():
                descriptions.append(
                    zope.lifecycleevent.Attributes(interface, *names))
            # Send out a detailed object-modified event
            zope.event.notify(
                zope.lifecycleevent.ObjectModifiedEvent(
                    content, *descriptions))
        return changes

    def doHandleApply(self, action):
        """Extract data and calls applyChanges."""
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return False
        changes = self.applyChanges(data)
        if changes:
            self.status = self.successMessage
        else:
            self.status = self.noChangesMessage
        return True

    def doBack(self, action):
        """Process back action and return True on success."""
        if self.handleApplyOnBack:
            return self.doHandleApply(action)
        return True

    def doNext(self, action):
        """Process next action and return True on success."""
        if self.handleApplyOnNext:
            return self.doHandleApply(action)
        return True

    def doComplete(self, action):
        """Process complete action and return True on success."""
        if self.handleApplyOnComplete:
            return self.doHandleApply(action)
        return True

    def update(self):
        # setup wizard actions
        self.wizard.update()
        if self.nextURL is not None:
            # abort and do redirect in render method
            return
        # update and execute step actions
        super().update()
        # execute wizard actions
        self.wizard.actions.execute()

    def render(self):
        # render content template
        if self.nextURL is not None:
            self.request.response.redirect(self.nextURL)
            return ''
        return super().render()

    def __repr__(self):
        return "<{} '{}'>".format(self.__class__.__name__, self.__name__)


class EditStep(Step):
    """Step with default save button action."""

    @button.buttonAndHandler(_('Apply'), name='apply')
    def handleApply(self, action):
        self.doHandleApply(action)


class UnavailableStep(Step):
    """A step that is not available.

    This class is particularly useful for turning off an adapter step.
    """

    available = False



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import os

from setuptools import find_packages
from setuptools import setup


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as f:
        return f.read()


setup(
    name='z3c.wizard',
    version='2.0',
    author="Roger Ineichen and the Zope Community",
    author_email="zope-dev@zope.dev",
    description="Wizard based on z3c.form for for Zope3",
    long_description='\n\n'.join([
        read('README.rst'),
        '.. contents::',
        read('src', 'z3c', 'wizard', 'README.rst'),
        read('src', 'z3c', 'wizard', 'zcml.rst'),
        read('CHANGES.rst'),
    ]),
    license="ZPL 2.1",
    keywords="zope zope3 z3c form wizard",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: Zope :: 3',
    ],
    url='https://github.com/zopefoundation/z3c.wizard',
    packages=find_packages('src'),
    include_package_data=True,
    package_dir={'': 'src'},
    namespace_packages=['z3c'],
    python_requires='>=3.7',
    extras_require=dict(
        test=[
            'z3c.macro',
            'zope.app.pagetemplate',
            'zope.app.testing',
            'zope.browserresource',
            'zope.publisher',
            'zope.testing',
            'zope.testrunner',
        ],
    ),
    install_requires=[
        'setuptools',
        'z3c.form >= 2.0',
        'z3c.formui',
        'z3c.pagelet',
        'zope.browserpage',
        'zope.component',
        'zope.configuration',
        'zope.event',
        'zope.i18nmessageid',
        'zope.interface',
        'zope.lifecycleevent',
        'zope.location',
        'zope.publisher',
        'zope.schema',
        'zope.security',
        'zope.traversing',
    ],
    zip_safe=False,
)



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Wizard button actions implementation
$Id:$
"""
__docformat__ = "reStructuredText"

from z3c.form import button

from z3c.wizard import interfaces


class WizardButtonActions(button.ButtonActions):
    """Wizard Button Actions."""

    @property
    def backActions(self):
        return [action for action in self.values()
                if interfaces.IBackButton.providedBy(action.field)]

    @property
    def forwardActions(self):
        return [action for action in self.values()
                if interfaces.INextButton.providedBy(action.field)]



##############################################################################
#
# Copyright (c) 2008-2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import z3c.macro.tales
import zope.i18n.interfaces
import zope.i18n.negotiator
from zope.app.pagetemplate import metaconfigure
from zope.app.testing import setup


###############################################################################
#
# testing setup
#
###############################################################################

def setUp(test):
    test.globs = {'root': setup.placefulSetUp(True)}

    zope.component.provideUtility(zope.i18n.negotiator.Negotiator(),
                                  zope.i18n.interfaces.INegotiator)
    metaconfigure.registerType('macro', z3c.macro.tales.MacroExpression)


def tearDown(test):
    setup.placefulTearDown()



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import zope.component
import zope.interface
from z3c.form import button
from z3c.formui import form
from zope.publisher.interfaces import NotFound
from zope.traversing.browser import absoluteURL

from z3c.wizard import interfaces
from z3c.wizard.button import WizardButtonActions


def nameStep(step, name):
    """Give a step a __name__."""
    step.__name__ = name
    return step


@zope.interface.implementer(interfaces.IWizard)
class Wizard(form.Form):
    """Wizard form.

    The wizard is responsible for manage the steps and offers the wizard menu
    navigation and knows the step order. The wizard can check the conditions
    given from the steps. The wizard is also responsible for delegate the
    back, next and complete actions to the steps.

    This IWizard object is modelled as a Controller known from the MVC
    (Model, view, controller) patter version 2.0 and the step is implemented as
    a view.
    """

    buttons = button.Buttons(interfaces.IWizardButtons)

    # customize this part if needed
    stepInterface = interfaces.IStep

    firstStepAsDefault = True
    adjustStep = True
    confirmationPageName = None
    nextURL = None

    cssActive = 'selected'
    cssInActive = None  # None will skip class attribute in DOM element

    # for internal use
    __name__ = None
    steps = None
    step = None

    @property
    def baseURL(self):
        return absoluteURL(self, self.request)

    def setUpSteps(self):
        """Return a list of steps. This implementation uses IStep adapters.

        Take a look at the addStep method defined in step.py. This method
        allows you to setup steps directly in the method and offers an API for
        customized step setup.
        """
        steps = list(zope.component.getAdapters(
            (self.context, self.request, self), self.stepInterface))
        return [nameStep(step, name) for name, step in steps]

    def filterSteps(self, steps):
        """Make sure to only select available steps and we give a name."""
        return [step for step in steps if step.available]

    def orderSteps(self, steps):
        # order steps by it's weight
        return sorted(steps, key=lambda step: step.weight)

    @property
    def steps(self):
        steps = self.setUpSteps()
        steps = self.filterSteps(steps)
        return self.orderSteps(steps)

    @property
    def completed(self):
        for step in self.steps:
            if not step.completed:
                return False
        return True

    @property
    def isFirstStep(self):
        """See interfaces.IWizard"""
        return self.step and self.step.__name__ == self.steps[0].__name__

    @property
    def isLastStep(self):
        """See interfaces.IWizard"""
        return self.step and self.step.__name__ == self.steps[-1].__name__

    @property
    def showBackButton(self):
        """Ask the step."""
        return self.step and self.step.showBackButton

    @property
    def showNextButton(self):
        """Ask the step."""
        return self.step and self.step.showNextButton

    @property
    def showCompleteButton(self):
        """Ask the step."""
        return self.step.showCompleteButton

    @property
    def previousStepName(self):
        if self.step is None:
            return
        stepNames = [step.__name__ for step in self.steps]
        idx = stepNames.index(self.step.__name__)
        if idx == 0:
            return
        return stepNames[idx - 1]

    @property
    def nextStepName(self):
        if self.step is None:
            return
        stepNames = [step.__name__ for step in self.steps]
        idx = stepNames.index(self.step.__name__)
        if idx == len(stepNames) - 1:
            return
        return stepNames[idx + 1]

    @property
    def stepMenu(self):
        items = []
        append = items.append
        lenght = len(self.steps) - 1
        for idx, step in enumerate(self.steps):
            firstStep = False
            lastStep = False
            if step.visible:
                isSelected = self.step and self.step.__name__ == step.__name__
                cssClass = isSelected and self.cssActive or self.cssInActive
                if idx == 0:
                    firstStep = True
                if idx == lenght:
                    lastStep = True
                append({
                    'name': step.__name__,
                    'title': step.label,
                    'number': str(idx + 1),
                    'url': '{}/{}'.format(self.baseURL, step.__name__),
                    'selected': self.step.__name__ == step.__name__,
                    'class': cssClass,
                    'first': firstStep,
                    'last': lastStep
                })
        return items

    def getDefaultStep(self):
        """Can return the first or first not completed step as default."""
        # return first step if this option is set
        if self.firstStepAsDefault:
            return self.steps[0]
        # return first not completed step
        for step in self.steps:
            if not step.completed:
                return step
        # fallback to first step if all steps completed
        return self.steps[0]

    def doAdjustStep(self):
        # Make sure all previous steps got completed. If not, redirect to the
        # last incomplete step
        if not self.adjustStep:
            return False
        for step in self.steps:
            if step.__name__ is self.step.__name__:
                break
            if not step.completed:
                # prepare redirect to not completed step and return True
                self.nextURL = '{}/{}'.format(self.baseURL, step.__name__)
                return True
        # or return False
        return False

    def updateActions(self):
        self.actions = WizardButtonActions(self, self.request, self.context)
        self.actions.update()

    def update(self):
        if self.doAdjustStep():
            return
        self.updateActions()

    def publishTraverse(self, request, name):
        """Traverse to step by it's name."""
        # Remove HTML ending
        if '.' in name:
            rawName = name.rsplit('.', 1)[0]
        else:
            rawName = name
        # Find the active step
        for step in self.steps:
            if step.__name__ == rawName:
                self.step = step
                return self.step
        raise NotFound(self, name, request)

    def browserDefault(self, request):
        """The default step is our browserDefault traversal setp."""
        if self.step is None:
            step = self.getDefaultStep()
        # always return default step as default view for our wizard
        return self, (step.__name__,)

    def goToStep(self, stepName):
        self.nextURL = '{}/{}'.format(self.baseURL, stepName)

    def goToBack(self):
        # redirect to next step if previous get sucessfuly processed
        self.goToStep(self.previousStepName)

    def goToNext(self):
        # redirect to next step if previous get sucessfuly processed
        self.goToStep(self.nextStepName)

    def doBack(self, action):
        if self.step.doBack(action):
            self.goToBack()

    def doNext(self, action):
        if self.step.doNext(action):
            self.goToNext()

    def doComplete(self, action):
        if self.step.doComplete(action):
            # do finsih after step get completed is completed
            self.doFinish()

    def doFinish(self):
        """Force redirect after doComplete if confirmationPageName is given."""
        if self.confirmationPageName is not None:
            self.nextURL = '{}/{}'.format(
                absoluteURL(self.context, self.request),
                self.confirmationPageName)

    @button.handler(interfaces.IWizardButtons['back'])
    def handleBack(self, action):
        self.doBack(action)

    @button.handler(interfaces.IWizardButtons['next'])
    def handleNext(self, action):
        self.doNext(action)

    @button.handler(interfaces.IWizardButtons['complete'])
    def handleComplete(self, action):
        self.doComplete(action)

    def render(self, *args, **kws):
        raise NotImplementedError('render is no supported')

    def __repr__(self):
        return "<{} '{}'>".format(self.__class__.__name__, self.__name__)
